{-# LANGUAGE InstanceSigs #-}
{-# LANGUAGE TypeOperators #-}
{-# LANGUAGE FlexibleContexts #-}
{-# LANGUAGE UndecidableInstances #-}
{-# LANGUAGE ScopedTypeVariables #-}
{-# LANGUAGE TypeApplications #-}
{-# LANGUAGE StandaloneDeriving #-}
{-# LANGUAGE DerivingVia #-}
{-# LANGUAGE FlexibleInstances #-}
{-# LANGUAGE DataKinds #-}
module Language.Java.Classfile.Extractable (Extractable(extract), expectConstant) where

import Language.Java.Classfile.Extract (Extract, bytes, expectEqual, traceType, traceConstructor, traceField, traceIndex)
import Data.Word (Word8, Word16, Word32)
import qualified Data.ByteString.Lazy as ByteString
import Data.ByteString.Lazy (ByteString)
import Data.Bits (Bits(shiftL, (.|.)))
import GHC.Generics (U1 (U1), (:*:) ((:*:)), (:+:) (L1, R1), M1 (M1), K1 (K1), Generic (Rep, to), Generically (Generically), Meta(..))
import Control.Applicative ((<|>))
import Data.Array.IArray (Array, listArray, Ix)
import Control.Monad (forM)
import Data.Int (Int32, Int64)
import GHC.Float (castWord32ToFloat, castWord64ToDouble)
import qualified Data.Text as Text
import GHC.TypeLits (symbolVal, KnownSymbol)
import Data.Proxy (Proxy(Proxy))
import Data.Typeable (Typeable)

class Extractable a where
  extract :: Extract a

instance Extractable Word8 where
  extract :: Extract Word8
  extract = ByteString.head <$> bytes 1

instance Extractable Word16 where
  extract :: Extract Word16
  extract = build <$> bytes 2

instance Extractable Word32 where
  extract :: Extract Word32
  extract = build <$> bytes 4

instance Extractable Int32 where
  extract :: Extract Int32
  extract = fromIntegral <$> extract @Word32

instance Extractable Float where
  extract :: Extract Float
  extract = castWord32ToFloat <$> extract

instance Extractable Double where
  extract :: Extract Double
  extract = castWord64ToDouble . build <$> bytes 8

instance Extractable Int64 where
  extract :: Extract Int64
  extract = do
    high <- extract @Word32
    low <- extract @Word32

    pure $ (fromIntegral high `shiftL` 32) .|. fromIntegral low

instance (Extractable index, Extractable element, Ix index, Integral index) => Extractable (Array index element) where
  extract :: Extract (Array index element)
  extract = do
    count <- extract
    elements <- forM [1..fromIntegral count - 1] $ \ i -> traceIndex i extract
    pure $ listArray (1, count - 1) elements

deriving via Generically () instance Extractable ()
deriving via Generically (a, b) instance (Extractable a, Extractable b) => Extractable (a, b)
deriving via Generically (a, b, c) instance (Extractable a, Extractable b, Extractable c) => Extractable (a, b, c)

expectConstant :: (Extractable a, Eq a, Show a, Typeable a) => a -> Extract a
expectConstant = expectEqual extract

build :: (Bits a, Num a) => ByteString -> a
build = ByteString.foldl' shiftOr 0

shiftOr :: (Bits a1, Num a1) => a1 -> Word8 -> a1
shiftOr word appendix = shiftL word 8 .|. fromIntegral appendix


class GenericExtractable a where
  genericExtract :: Extract (a i)

-- sum types
instance (GenericExtractable l, GenericExtractable r) => GenericExtractable (l :+: r) where
  genericExtract :: Extract ((:+:) l r i)
  genericExtract = (L1 <$> genericExtract) <|> (R1 <$> genericExtract)

-- unit data types
instance GenericExtractable U1 where
  genericExtract :: Extract (U1 i)
  genericExtract = pure U1

-- product types
instance (GenericExtractable l, GenericExtractable r) => GenericExtractable (l :*: r) where
  genericExtract :: Extract ((:*:) l r i)
  genericExtract = liftA2 (:*:) genericExtract genericExtract

-- meta information constructors
instance (GenericExtractable a, KnownSymbol name) => GenericExtractable (M1 tag (MetaData name module_ package isNewtype) a) where
  genericExtract :: Extract (M1 tag (MetaData name module_ package isNewtype) a i)
  genericExtract = M1 <$> traceType typeName genericExtract
    where
      typeName = Text.pack $ symbolVal (Proxy @name)

instance (GenericExtractable a, KnownSymbol name) => GenericExtractable (M1 tag (MetaCons name fixity isRecord) a) where
  genericExtract :: Extract (M1 tag (MetaCons name fixity isRecord) a i)
  genericExtract = M1 <$> traceConstructor conName genericExtract
    where
      conName = Text.pack $ symbolVal (Proxy @name)

instance (GenericExtractable a, KnownSymbol name) => GenericExtractable (M1 tag (MetaSel (Just name) isUnpacked isStrictSource isStrictCompiler) a) where
  genericExtract = M1 <$> traceField fieldName genericExtract
    where
      fieldName = Text.pack $ symbolVal (Proxy @name)

instance (GenericExtractable a) => GenericExtractable (M1 tag (MetaSel Nothing isUnpacked isStrictSource isStrictCompiler) a) where
  genericExtract = M1 <$> genericExtract

-- fields !! yay
instance Extractable a => GenericExtractable (K1 i a) where
  genericExtract :: Extract (K1 i1 a i2)
  genericExtract = K1 <$> extract

instance (Generic a, GenericExtractable (Rep a)) => Extractable (Generically a) where
  extract :: Extract (Generically a)
  extract = Generically . to <$> genericExtract

